#!D:\bld\refnx_1580789782956\_h_env\pythonw.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'refnx==0.1.11','gui_scripts','slim'
__requires__ = 'refnx==0.1.11'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('refnx==0.1.11', 'gui_scripts', 'slim')()
    )
